#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include "typedefs.h"


/*
This function will create a new node and insert it AFTER the
InsertPoint in the list of nodes.

Parameters:
  InsertPoint - pointer to the node after which the new node
                will be inserted
  Type        - Type of the new node (TEXT or any of the supported tags)
  TextString  - the text string for the new node,
                in case the new node is required to be of type TEXT

Note:
  InsertPoint should not be NULL.

Working:
  This function will first allocate memory for a new node; if this fails,
  it will return NULL.
  Then it will update the 'type' field of the new node and insert the
  new node after InsertPoint.
  In case, the type of the new node is TEXT, the 's' field of the new
  node will be set to the input parameter 'TextString'. Then it will look
  for the nearest text-type node in both forward and backward directions
  and update the linked list of text-type nodes too. The 'match' field
  of the new node will be set to NULL.
  If the type of the new node is not TEXT (i.e. it is a tag), the
  's', 'prevtext' & 'nexttext' fields of the new node are set to NULL.
  However the 'match' field is NOT updated.
  The 'in' field of the new node i.e. its incoming context is NEVER
  updated.

Limitations:
  This function cannot be used to insert a node at the beginning
  of the list i.e before the root node.

*/

NODE *List_InsertAfterNode( NODE *InsertPoint, int Type, char *TextString )
{
    NODE *NewNode;
    NODE *p, *n;

    NewNode = newNode();

    if ( NewNode == NULL )
    {
        return NULL;
    }

    NewNode->type = Type;

    NewNode->prev = InsertPoint;
    NewNode->next = InsertPoint->next;

    InsertPoint->next = NewNode;
    if ( NewNode->next != NULL )  /* check for last node */
    {
        NewNode->next->prev = NewNode;
    }

    if ( Type == TEXT )
    {
        NewNode->s = TextString;

        p = NewNode->prev;

        /* Scan backwards for the nearest TEXT node */
        while ( p != NULL )
        {
            if ( p->type == TEXT )
            {
                break;
            }
            p = p->prev;
        }

        /* If found, we also get the nearest TEXT node in forward direction */
        if ( p != NULL )
        {
            n = p->nexttext;
        }
        else
        {
            /* otherwise, look forward for the nearest TEXT node */
            n = NewNode->next;

            while ( n != NULL )
            {
                if ( n->type == TEXT )
                {
                    break;
                }
                n = n->next;
            }
        }

        NewNode->prevtext = p;
        NewNode->nexttext = n;

        if ( p != NULL )   /* check for first text node */
        {
            p->nexttext = NewNode;
        }

        if ( n != NULL )   /* check for last text node */
        {
            n->prevtext = NewNode;
        }

        NewNode->match = NULL;
    }
    else
    {
        /* for non-text nodes */
        NewNode->s = NULL;
        NewNode->prevtext = NULL;
        NewNode->nexttext = NULL;
    }

    return NewNode;
}

/*
Given a node, which is a 'starting tag' node, this function will
recompute the incoming contexts of every node from the given node
(inclusive) till the matching 'ending tag' node (inclusive).

Parameters:
  StartNode - the node from which to start; this must necessarily
              be a 'starting tag' type node

Notes:
  Cannot be used for a text-type node

*/

void AdjustContext( NODE *StartNode )
{
    NODE *p, *q;

    if ( !isstarttag( StartNode->type ) )
    {
        return;
    }

    p = StartNode;
    q = StartNode->match;

    if ( q == NULL )
    {
        return;
    }

    while ( p != q )
    {
        if ( p->prev != NULL )
        {
            /*
            if the type of the previous node is an ending tag, the incoming context for this node,
            is the same as the incoming context, for the matching opening tag for that ending tag
            */
            if ( isendtag( p->prev->type ) )
            {
                p->in = p->prev->match->in;
            }
            else /* previous is an opening tag or text */
            {
                p->in = AddTagToContext(&p->prev->in, p->prev);
            }
        }
        else
        {
            /* This is the first node */
        }

        p = p->next;
    }

    /*
    Above loop comes out at the matching ending node, whose context is not updated.
    Hence repeat the action for this node
    */

    if ( isendtag( p->prev->type ) )
    {
        p->in = p->prev->match->in;
    }
    else
    {
        p->in = AddTagToContext(&p->prev->in, p->prev);
    }
}

/* Checks whether the 2 given contexts are equal - without regard to colour & size */
int IsNonClrNonSizeCtxtEql( CONTEXT *c1, CONTEXT *c2 )
{
    return ( ( c1->b  == c2->b  )  &&
             ( c1->em == c2->em )  &&
             ( c1->i  == c2->i  )  &&
             ( c1->s  == c2->s  )  &&
             ( c1->tt == c2->tt )  &&
             ( c1->u  == c2->u  )
           );
}

/* Checks whether the given context satisfies the condition of plainness */
int IsPlainTextContext( CONTEXT *c )
{
    return ( ( c->b  == 0 )  &&
             ( c->em == 0 )  &&
             ( c->i  == 0 )  &&
             ( c->s  == 0 )  &&
             ( c->tt == 0 )  &&
             ( c->u  == 0 )
           );
}

/* Gives a measure of how far we are from a plainness context */
int NonPlainCount( CONTEXT *c )
{
    int count = 0;

    if ( c->b  ) count++;
    if ( c->em ) count++;
    if ( c->i  ) count++;
    if ( c->s  ) count++;
    if ( c->tt ) count++;
    count += c->u;

    return count;
}

int Is_b_em_i_s_tt_u_StartTag( int Type )
{
    return ( ( Type >= BOLD ) && ( Type <= UNDERLINE ) && ( Type != PLAIN ) );
}

int Is_b_em_i_s_tt_u_EndTag( int Type )
{
    return ( ( Type >= BOLD_END ) && ( Type <= UNDERLINE_END ) && ( Type != PLAIN_END ) );
}

NODE *PL_ShortCut( NODE *root )
{
    NODE *node, *TextNode;
    NODE *PrevNonPlainText, *NextNonPlainText;
    NODE *node1, *node2;
    NODE *node3, *node4;
    NODE *node5, *node6;
    NODE *AdjustStartNode;
    NODE *arr1[10], *arr2[10];
    int count1, count2;
    int i, j;
    int PLTagFound;
    int SameSeq;
    int NoTextInBetween;
    int OkToContinue;

    node = root;
    TextNode = NULL;

    /* Search for a text node */
    while ( node != NULL )
    {
        if ( node->type == TEXT )
        {
            TextNode = node;
            break;
        }
        node = node->next;
    }

    /* Search for a Plain Text node among all the text nodes */
    /* This is the main search loop */
    while ( TextNode != NULL )
    {
        if ( !IsPlainTextContext( &TextNode->in ) )
        {
            TextNode = TextNode->nexttext;
            continue;
        }

        /* We have found a plain text at TextNode now */

        NextNonPlainText = TextNode->nexttext;

        /* Search for the next NonPlainText */
        while ( NextNonPlainText != NULL )
        {
            if ( !IsPlainTextContext( &NextNonPlainText->in ) )
            {
                break;
            }
            NextNonPlainText = NextNonPlainText->nexttext;
        }

        /* If there isn't any NonPlainText after this plain text */
        /* we quit the main search loop */
        if ( NextNonPlainText == NULL )
        {
            break;
        }

        /* The previous NonPlainText, if it exists is the previous text node */
        PrevNonPlainText = TextNode->prevtext;

        /* If there isn't any NonPlainText before this plain text
           we continue the main search loop from NextNonPlainText */
        if ( PrevNonPlainText == NULL )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        /* If PrevNonPlainText & NextNonPlainText do not have the
           same context (disregarding size and colour)
           we continue the main search loop from NextNonPlainText */
        if ( !IsNonClrNonSizeCtxtEql( &PrevNonPlainText->in, &NextNonPlainText->in ) )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        /* If the NonPlainText is only one tag away from a plainness
           context, we do not achieve any compression by PL shortcut;
           we continue the main search loop from NextNonPlainText */
        if ( NonPlainCount( &PrevNonPlainText->in ) <= 1 )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        /*
        We now make an IMPORTANT ASSUMPTION viz. that the redundancy
        and the dependency passes have run before this function is
        called - so that all tags in the list which are redundant and
        which do not have any effect on any text node have been
        removed.
        Under this assumption, the following 2 statements hold:
        1. The tags between the PrevNonPlainText and the plain text
        at TextNode are arranged such that all the ending tags precede
        the starting tags.
        2. Similarly the tags between a) the text node (necessarily a plain
        text) immediately previous to NextNonPlainText and b) NextNonPlainText
        are again arranged such that all the ending tags precede the starting
        tags
        */

        node1 = TextNode->prev;
        PLTagFound = 0;

        /* Search for a PLAIN tag among the starting tags prior to the Plain Text */
        while ( 1 )
        {
            if ( node1->type == PLAIN )
            {
                PLTagFound = 1;
                break;
            }

            if ( !isstarttag( node1->type ) )
            {
                break;
            }

            node1 = node1->prev;
        }

        /* If a PLAIN tag is found, the plain text is of no use to us.
        we continue the main search loop from NextNonPlainText */

        if ( PLTagFound )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        /* Note that now node1 is the point after which we can insert a PLAIN tag */
        node3 = node1;
        i = 0;

        /*
        Put the ending tags (non-colour, non-size, non-PLAIN) following PrevNonPlainText
        (these are the tags that we propose to delete) in arr1, but in the reverse order
        */

        while ( 1 )
        {
            if ( !isendtag( node3->type ) )
            {
                break;
            }

            if ( Is_b_em_i_s_tt_u_EndTag( node3->type ) )
            {
                arr1[i++] = node3;
            }

            node3 = node3->prev;
        }

        count1 = i;

        node2 = NextNonPlainText->prev;
        i = 0;

        /*
        Similarly put the starting tags (non-colour, non-size, non-PLAIN) preceding NextNonPlainText
        (these are the tags that we propose to delete) in arr2, but in the reverse order
        */
        while ( 1 )
        {
            if ( !isstarttag( node2->type ) )
            {
                break;
            }

            if ( Is_b_em_i_s_tt_u_StartTag( node2->type ) )
            {
                arr2[i++] = node2;
            }

            node2 = node2->prev;
        }

        /* Note that now node2 is the point after which we can insert a PLAIN_END tag */

        count2 = i;

        /* The 2 counts should always match - but just in case they don't,
        we continue the main search loop from NextNonPlainText */

        if ( ( count1 != count2 ) || ( count1 == 0 ) )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        i = 0;
        j = count1 - 1;
        SameSeq = 1;

        /* Check whether the sequence of ending tags is the same as the sequence of starting tags */
        while ( count2 )
        {
            if ( arr1[i++]->type != arr2[j--]->type)
            {
                SameSeq = 0;
                break;
            }
            count2--;
        }

        if ( !SameSeq )
        {
            /*
            If the sequences are not the same, then we must verify that there is no text
            in between the sequence of matching tags corresponding to the sequence of
            starting tags - only then can we proceed with PL shortcut
            */

            j = 0;
            NoTextInBetween = 1;

            while ( j < count1 - 1 )
            {
                node5 = arr2[j]->match;
                node6 = arr2[j+1]->match;

                node5 = node5->next;

                while ( node5 != node6 )
                {
                    if ( node5->type == TEXT )
                    {
                        NoTextInBetween = 0;
                        break;
                    }
                    node5 = node5->next;
                }

                if ( !NoTextInBetween )
                {
                    break;
                }

                j++;
            }

            if ( NoTextInBetween )
            {
                OkToContinue = 1;
            }
            else
            {
                OkToContinue = 0;
            }
        }
        else
        {
            /* If the sequences are the same, then we can always proceed with PL shortcut */
            OkToContinue = 1;
        }

        /* If we cannot proceed with PL shortcut
        we continue the main search loop from NextNonPlainText */
        if ( !OkToContinue )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        /* Insert the PLAIN and PLAIN_END tags */
        node5 = List_InsertAfterNode( node1, PLAIN, NULL );
        if ( node5 == NULL )
        {
            TextNode = NextNonPlainText;
            continue;
        }

        node6 = List_InsertAfterNode( node2, PLAIN_END, NULL );
        if ( node6 == NULL )
        {
            List_remove( root, node5 );
            TextNode = NextNonPlainText;
            continue;
        }

        /* Match the 2 inserted nodes */
        node5->match = node6;
        node6->match = node5;

        /* The following is the point from which we need to adjust/update the list
        after we are done */
        AdjustStartNode = arr1[0]->match;

        i = count1 - 1;
        j = 0;

        /*
        Now remove the nodes from each sequence a pair at a time, after
        a. setting the new type for the matching tag of each of the starting tag
           that precedes NextNextNonPlainText
        b. Matching the matching tags of the 2 sequences
        */

        while ( j < count1 )
        {
            node4 = arr1[i--];
            node5 = arr2[j++];

            node3 = node4->match;
            node6 = node5->match;

            node6->type = node4->type;

            node6->match = node3;
            node3->match = node6;

            List_remove( root, node4 );
            List_remove( root, node5 );
        }

        AdjustContext( AdjustStartNode );

        TextNode = NextNonPlainText;
    }

    return root;
}
